/* 
THREE.js r101 
*/

// Global Variables
let canvas = document.getElementById("myCanvas");
let camera0, scene0, renderer, controls, clock, stats;
let composer, glowComposer, glowRenderTarget, additivePass, glowScene, glowObjects = [];
let textureLoader;
let Textures = {};
let Lights = [];
let shadows = true;

let additivePassShader = {
	defines: {},
	
	uniforms: {
		tDiffuse: { type: "t", value: null },
		tGlow: { type: "t", value: null },
	},
	
	vertexShader: `
		varying vec2 vUv;
		
		void main(){
			vUv = uv;
			gl_Position = projectionMatrix * modelViewMatrix * vec4( position , 1.0 );
		}
	`,
	
	fragmentShader: `
		
		uniform sampler2D tDiffuse;
		uniform sampler2D tGlow;
		
		varying vec2 vUv;
		
		void main(){
			vec4 color = texture2D( tDiffuse , vUv );
			vec4 glow = texture2D( tGlow , vUv );
			vec4 mixed = color + glow;
			
			gl_FragColor = glow;
		}
	`,
};

function init() {
	// Renderer
	renderer = new THREE.WebGLRenderer( { canvas: canvas, antialias: false } );
	renderer.setSize( window.innerWidth, window.innerHeight );
	if(shadows){ 
		renderer.shadowMap.enabled = true;
		renderer.shadowMap.type = THREE.PCFSoftShadowMap;
		renderer.shadowMap.autoUpdate = false;
	}
	
	// Scene
	scene0 = new THREE.Scene();
	scene0.name = 'scene0';
	scene0.background = new THREE.Color( 0x101010 );
	
	// Camera
	camera0 = new THREE.PerspectiveCamera( 60, window.innerWidth / window.innerHeight, 0.1, 1000 );
	camera0.position.set( 0 , 3 , 10 );
	
	// Clock
	clock = new THREE.Clock();
	
	//Stats
	stats = new Stats();
	document.body.appendChild( stats.dom );
	
	// Loaders
	textureLoader = new THREE.TextureLoader();

	// Resize Event
	window.addEventListener("resize", function(){
		renderer.setSize( window.innerWidth, window.innerHeight );
		camera0.aspect = window.innerWidth / window.innerHeight;
		camera0.updateProjectionMatrix();
	}, false);
	
	// Inits
	initControls();
	
	initLights();
	createStartingMesh();
	initPostProcessing();
}

let createStartingMesh = function(){
	let floor = new THREE.Mesh(
		new THREE.PlaneBufferGeometry( 30 , 30 ),
		new THREE.MeshPhongMaterial({
			color: 0x404040,
			shininess: 0,
		})
	);
	floor.rotation.x -= 90 * Math.PI/180;
	scene0.add( floor );
	if(shadows) floor.receiveShadow = true;
	
	let cube = new THREE.Mesh( 
		new THREE.CylinderBufferGeometry( 1 , 1 , 2 , 20 ) , 
		new THREE.MeshPhongMaterial({color: 0x8080a0 })
	);
	if(shadows) {
		cube.castShadow = true;
		cube.receiveShadow = true;
	}
	cube.position.set( 0 , 1 , 0 );
	scene0.add( cube );
	glowObjects.push( cube );
}

let initControls = function(){
	controls = new THREE.OrbitControls( camera0 )
}

let initLights = function(){
	Lights[0] = new THREE.AmbientLight( 0xffffff , 0.3 );
	Lights[1] = new THREE.PointLight( 0xffffff , 1.2 );
	Lights[1].position.set( -3 , 5 , 4 );
	if(shadows){
		Lights[1].castShadow = true;
		Lights[1].shadow.mapSize.width = 1024 * 2;
		Lights[1].shadow.mapSize.height = 1024 * 2;
		Lights[1].shadow.camera.near = 0.1;
		Lights[1].shadow.camera.far = 50;
		if( Lights[1] instanceof THREE.DirectionalLight ){
			Lights[1].shadow.camera.left = -30;
			Lights[1].shadow.camera.bottom = -30;
			Lights[1].shadow.camera.top = 30;
			Lights[1].shadow.camera.right = 30;
		}
		// Lights[1].shadow.bias = 0.0001;
	}
	let helper = new THREE.Mesh( 
		new THREE.SphereBufferGeometry(0.2) , 
		new THREE.MeshBasicMaterial()
	);
	helper.position.copy( Lights[1].position );
	scene0.add( helper );
	
	for(let i = 0; i < Lights.length; i++){
		scene0.add( Lights[i] );
	}
}

let initPostProcessing = function(){
	
	initGlowScene( scene0 );
	
	composer = new THREE.EffectComposer( renderer );
	
	let glowRenderer = new THREE.WebGLRenderer({ canvas: canvas });
	glowRenderTarget = new THREE.WebGLRenderTarget( 512 , 512 , {} );
	glowComposer = new THREE.EffectComposer( glowRenderer );
	
	let renderPass = new THREE.RenderPass( scene0 , camera0 );
	let fxaaPass = new THREE.ShaderPass( THREE.FXAAShader );
	additivePass = new THREE.ShaderPass( additivePassShader );
	
	let glowPass = new THREE.RenderPass( glowScene , camera0 );
	let bluriness = 3;
	let hBlur = new THREE.ShaderPass( THREE.HorizontalBlurShader );
	hBlur.uniforms.h.value = bluriness/canvas.height;
	let vBlur = new THREE.ShaderPass( THREE.VerticalBlurShader );
	vBlur.uniforms.v.value = bluriness/canvas.width;
	
	glowComposer.addPass( glowPass );
	glowComposer.addPass( hBlur );
	glowComposer.addPass( vBlur );
	// vBlur.renderToScreen = true;
	
	composer.addPass( renderPass );
	composer.addPass( fxaaPass );
	composer.addPass( additivePass );
	additivePass.renderToScreen = true;
	
	console.log( glowComposer , composer );
}

let initGlowScene = function( scene ){
	
	glowScene = new THREE.Scene();
	glowScene.name = 'glowScene';
	
	const blackMaterial = new THREE.MeshBasicMaterial({ color: 0x000000 });
	const glowMaterial = new THREE.MeshBasicMaterial({ color: 0xFFFFFF });
	scene.children.forEach( function( child ){
		
		let clone = child.clone();
		if( glowObjects.indexOf( child ) != -1 ) { // is in array
			clone.material = glowMaterial;
		} else {
			clone.material = blackMaterial;
		}
		
		glowScene.add( clone );
	});
}

function animate() {
	stats.begin();
	let delta = clock.getDelta();
	
	if( shadows ) renderer.shadowMap.needsUpdate = true;
	
	glowComposer.render( glowScene , camera0 );
	additivePass.uniforms.tGlow.value = glowRenderTarget.texture;
	
	composer.render( scene0, camera0 );
	
	// renderer.render( glowScene, camera0 );
	
	requestAnimationFrame( animate );
	stats.end();
}

init();
requestAnimationFrame( animate );
